import os
from numpy import gradient

# torch imports
import torch
from transformers import DetrForObjectDetection
import lightning as l
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.loggers import CSVLogger

# local imports
from niche_cv.detection.data.loaders.coco import CocoDetectionDataModule

# Reference:
# https://github.com/NielsRogge/Transformers-Tutorials/blob/master/DETR/Fine_tuning_DetrForObjectDetection_on_custom_dataset_(balloon).ipynb


class Niche_Detr:
    def __init__(self, path_model, dir_out, name_task, device):
        """
        Folder structure
        models/
            path to a directory containing outputs of save_pretrained() or HuggingFace model name
        out/
            train/
                <name_task>/
                    weights/
                        best.pt
                    ...
            val/
                <name_task>/
                    results.json
                    ...

        """
        # attributes
        self.model = None
        self.path_model = path_model
        self.dir_train = os.path.join(dir_out, "train")
        self.dir_val = os.path.join(dir_out, "val")
        self.name_task = name_task
        self.device = device

    def train(self, path_data, batch=16, epochs=100):
        # prepare dataset
        data_module = CocoDetectionDataModule(path_data, batch_size=batch)
        data_module.setup()
        id2label = {k: v["name"] for k, v in data_module.train.coco.cats.items()}

        # conifg model
        self.model = Detr(self.path_model, len(id2label))
        dir_out = create_output_dir(self.dir_train, self.name_task)
        callbacks = config_callbacks(dir_out)
        logger = config_loggers(dir_out)

        # setup trainer
        trainer = l.Trainer(
            max_epochs=epochs,
            gradient_clip_val=0.1,
            callbacks=callbacks,
            logger=logger,
        )
        trainer.fit(self.model, data_module)


def create_output_dir(path_root, name_task):
    """
    Create the output directory if it does not exist.
    """
    path_project = os.path.join(path_root, "out", "train", name_task)
    if not os.path.exists(path_project):
        os.makedirs(path_project)

    return path_project


def config_callbacks(dir_out):
    checkpoint_callback = ModelCheckpoint(
        monitor="val_loss",
        mode="min",
        save_top_k=1,
        verbose=False,
        save_last=False,
        dirpath=dir_out,
        # filename="checkpoint-{val_loss:.6f}",
        filename="model",
    )
    return [checkpoint_callback]


def config_loggers(dir_out):
    csv_logger = CSVLogger(
        save_dir=dir_out,
        name=".",  # won't create a new folder metrics
        version=".",  # won't create a new folder version_x
    )
    return [csv_logger]


class Detr(l.LightningModule):
    def __init__(
        self, backbone, n_labels, lr=1e-4, lr_backbone=1e-5, weight_decay=1e-4
    ):
        """
        backbone: e.g. "facebook/detr-resnet-50"
        n_labels: number of labels
        """
        super().__init__()
        # replace COCO classification head with custom head
        # we specify the "no_timm" variant here to not rely on the timm library
        # for the convolutional backbone
        self.model = DetrForObjectDetection.from_pretrained(
            backbone,
            revision="no_timm",
            num_labels=n_labels,
            ignore_mismatched_sizes=True,
        )
        # see https://github.com/PyTorchLightning/pytorch-lightning/pull/1896
        self.lr = lr
        self.lr_backbone = lr_backbone
        self.weight_decay = weight_decay

    def forward(self, pixel_values, pixel_mask):
        outputs = self.model(pixel_values=pixel_values, pixel_mask=pixel_mask)

        return outputs

    def common_step(self, batch, batch_idx):
        pixel_values = batch["pixel_values"]
        pixel_mask = batch["pixel_mask"]
        labels = [{k: v.to(self.device) for k, v in t.items()} for t in batch["labels"]]

        outputs = self.model(
            pixel_values=pixel_values, pixel_mask=pixel_mask, labels=labels
        )

        loss = outputs.loss
        loss_dict = outputs.loss_dict

        return loss, loss_dict

    def training_step(self, batch, batch_idx):
        loss, loss_dict = self.common_step(batch, batch_idx)
        # logs metrics for each training_step,
        # and the average across the epoch
        self.log("training_loss", loss)
        for k, v in loss_dict.items():
            self.log("train_" + k, v.item())

        return loss

    def validation_step(self, batch, batch_idx):
        loss, loss_dict = self.common_step(batch, batch_idx)
        self.log("validation_loss", loss)
        for k, v in loss_dict.items():
            self.log("validation_" + k, v.item())

        return loss

    def configure_optimizers(self):
        param_dicts = [
            {
                "params": [
                    p
                    for n, p in self.named_parameters()
                    if "backbone" not in n and p.requires_grad
                ]
            },
            {
                "params": [
                    p
                    for n, p in self.named_parameters()
                    if "backbone" in n and p.requires_grad
                ],
                "lr": self.lr_backbone,
            },
        ]
        optimizer = torch.optim.AdamW(
            param_dicts, lr=self.lr, weight_decay=self.weight_decay
        )

        return optimizer



"""
Folder structure:

root/
    test.txt
    images/
        img_1_13_jpg.rf.d69528304f2d10b633c6d94982185cb2.jpg
        img_2_13_jpg.rf.d69528304f2d10b633c6d94982185cb2.jpg
        ...
    labels/
        img_1_13_jpg.rf.d69528304f2d10b633c6d94982185cb2.txt
        img_2_13_jpg.rf.d69528304f2d10b633c6d94982185cb2.txt
        ...
    <suffix>/
        images/ # copy from root/images/
        labels/ # copy from root/labels/
        train.txt
        val.txt
        test.txt
        data.yaml

"""
## TODO: Add multi-threading vs. single-threading option

import os
import shutil
import random
import pathlib

from .splitter import Splitter


class YOLO_Splitter(Splitter):
    def __init__(
        self,
        path_root: str,
        classes: list,
        suffix: str = "",
        ratio_train: float = 0.64,
        ratio_val: float = 0.16,
        ratio_test: float = 0.2,
    ):
        self.suffix = suffix
        self.classes = classes
        self.path_yaml = None
        # copy images and labels
        if not os.path.exists(os.path.join(path_root, suffix, "images")):
            pathlib.Path(os.path.join(path_root, suffix)).mkdir(
                parents=True, exist_ok=True
            )
            shutil.copytree(
                os.path.join(path_root, "images"),
                os.path.join(path_root, suffix, "images"),
            )
            shutil.copytree(
                os.path.join(path_root, "labels"),
                os.path.join(path_root, suffix, "labels"),
            )
            shutil.copy(
                os.path.join(path_root, "test.txt"),
                os.path.join(path_root, suffix, "test.txt"),
            )
        self.path_root = os.path.join(path_root, suffix)
        super().__init__(self.path_root, ratio_train, ratio_val, ratio_test)

    def read_dataset(self):
        self.path_yaml = os.path.join(self.path_root, "data.yaml")
        f = open(self.path_yaml, "w")
        return f

    def get_ids(self):
        # from img_1_13_jpg.rf.jpg
        # to img_1_13_jpg.rf
        ls_imagefiles = os.listdir(os.path.join(self.path_root, "images"))
        ls_ids = [f[:-4] for f in ls_imagefiles if f.endswith(".jpg")]
        return ls_ids

    def shuffle_train_test(self):
        random.shuffle(self.ids)
        n = len(self.ids)
        n_train = int(self.ratio_train * n)
        n_val = int(self.ratio_val * n)
        # assignment
        self.id_train = self.ids[:n_train]
        self.id_val = self.ids[n_train : n_train + n_val]
        self.id_test = self.ids[n_train + n_val :]

    def shuffle_train_val(self, n_included=0):
        """
        prerequisite
        ------------
            test folder exists
        args
        ----
            n_included: int
                size of the training set and validation set. If 0, then use the entire available data.
        """
        # get test ids from txt
        path_txt = os.path.join(self.path_root, "test.txt")
        with open(path_txt, "r") as f:
            id_test = f.read().splitlines()
        id_test = [i[:-4] for i in id_test]  # 001.jpg -> 001
        # obtain which ids are to be shuffled
        id_remaining = [f for f in self.ids if f not in id_test]
        random.shuffle(id_remaining)
        # obtain each split size
        n_included = len(id_remaining) if n_included == 0 else n_included
        ratio_train = self.ratio_train / (self.ratio_train + self.ratio_val)
        n_train = int(ratio_train * n_included)
        n_val = n_included - n_train
        # assignment
        self.id_train = id_remaining[:n_train]
        self.id_val = id_remaining[n_train : n_train + n_val]
        self.id_test = id_test

    def write_dataset(self):
        self._write_yaml(classes=self.classes)
        for split in ["train", "val", "test"]:
            self._write_txt(split=split, ids=getattr(self, f"id_{split}"))
        return self.path_yaml

    def _write_yaml(self, classes: list):
        self.config.write(
            f"""
                path: {self.path_root}
                train: "train.txt"
                val: "val.txt"
                test: "test.txt"
                nc: {len(classes)}
                names: {classes}
            """
        )
        self.config.close()

    def _write_txt(self, split: str, ids: list):
        path_txt = os.path.join(self.path_root, "%s.txt" % split)
        with open(path_txt, "w") as f:
            for id in ids:
                f.write(os.path.join(self.path_root, "images", f"{id}.jpg") + "\n")

    # def _handle_folders(self):
    #     for s in ["train", "val", "test"]:
    #         dir_out = os.path.join(self.path_root, s)
    #         if os.path.exists(dir_out):
    #             shutil.rmtree(dir_out)
    #         os.mkdir(dir_out)
    #         os.mkdir(os.path.join(dir_out, "images"))
    #         os.mkdir(os.path.join(dir_out, "labels"))

    # def _copy_images_labels(self):
    #     for s in ["train", "val", "test"]:
    #         for id in getattr(self, f"id_{s}"):
    #             # copy images
    #             path_img = os.path.join(self.path_root, "images", f"{id}.jpg")
    #             path_img_out = os.path.join(self.path_root, s, "images", f"{id}.jpg")
    #             shutil.copy(path_img, path_img_out)
    #             # copy labels
    #             path_label = os.path.join(self.path_root, "labels", f"{id}.txt")
    #             path_label_out = os.path.join(self.path_root, s, "labels", f"{id}.txt")
    #             shutil.copy(path_label, path_label_out)



"""
Folder structure:

/root
    images/
        img_1.jpg
        img_2.jpg
        ...
    labels/
        _coco.json       <input>
        _coco_train.json <output>
        _coco_val.json   <output>
        _coco_test.json  <output>
        ...
"""

import os
import json
import random

from .splitter import Splitter


class CocoSplitter(Splitter):
    def __init__(
        self,
        root: str,  # root directory of the dataset, e.g. niche_cv/detection/data/balloon
        ratio_train: float = 0.64,
        ratio_val: float = 0.16,
        ratio_test: float = 0.2,
    ):
        super().__init__(root, ratio_train, ratio_val, ratio_test)
        self.dir_labels = os.path.join(self.root, "labels")
        self._set_dataset()

    def shuffle_train_test(self):
        """
        shuffle all ids into lists of train, val, and test
        """
        random.shuffle(self.ids)
        n = len(self.ids)
        n_train = int(self.ratio_train * n)
        n_val = int(self.ratio_val * n)
        # assignment
        self.id_train = self.ids[:n_train]
        self.id_val = self.ids[n_train : n_train + n_val]
        self.id_test = self.ids[n_train + n_val :]

    def write_dataset(self, name_json: str = "_coco"):
        dict_id = dict(
            train=self.id_train,
            val=self.id_val,
            test=self.id_test,
        )
        for split in dict_id:
            json_out = make_json(image_ids=dict_id[split], coco_annot=self.config)
            path_out = os.path.join(
                self.dir_labels, "{}_{}.json".format(name_json, split)
            )
            with open(path_out, "w") as f:
                json.dump(json_out, f)

    def _get_ids(self):
        return [img["id"] for img in self.config["images"]]

    def _set_dataset(self, name_json: str = "_coco"):
        """
        name_json: str
            name of the json file, e.g. _coco
        """
        # read or open the config file
        path_json = os.path.join(self.dir_labels, name_json + ".json")
        with open(path_json, "r") as f:
            coco_annot = json.load(f)

        # set ids from the config file
        self.config = coco_annot
        self._set_ids()


def make_json(image_ids: list, coco_annot: dict) -> dict:
    json_out = dict(
        info=coco_annot["info"],
        licenses=coco_annot["licenses"],
        images=[],
        annotations=[],
        categories=coco_annot["categories"],
    )
    for img in coco_annot["images"]:
        if img["id"] in image_ids:
            json_out["images"].append(img)
    for ann in coco_annot["annotations"]:
        if ann["image_id"] in image_ids:
            json_out["annotations"].append(ann)
    return json_out



import os
import json
from ultralytics import YOLO


class Niche_YOLO:
    def __init__(self, path_model, dir_out, name_task, device):
        """
        Folder structure
        models/
            <path_model>.pt
        out/
            train/
                <name_task>/
                    weights/
                        best.pt
                    ...
            val/
                <name_task>/
                    results.json
                    ...

        """
        # attributes
        self.model = None
        self.path_model = path_model
        self.dir_train = os.path.join(dir_out, "train")
        self.dir_val = os.path.join(dir_out, "val")
        self.name_task = name_task  # task folder name
        self.device = device

        # init
        self.load(path_model)

    def load(self, path_model):
        self.model = YOLO(path_model)
        print("model %s loaded" % path_model)

    def train(self, path_data, batch=16, epochs=100):
        """
        args
        ----
            path_data: str
                path to data.yaml
            batch: int
                batch size
            epochs: int
                number of epochs
        """
        self.model.train(
            data=path_data,
            batch=batch,
            epochs=epochs,
            device=self.device,
            project=self.dir_train,
            name=self.name_task,
            exist_ok=True,
        )
        best_model = os.path.join(self.dir_train, self.name_task, "weights", "best.pt")
        self.load(best_model)

    def evaluate(self, split="test"):
        metrics = self.model.val(
            split=split,
            device=self.device,
            project=self.dir_val,
            name=self.name_task,
            exist_ok=True,
        )
        # save metrics
        path_out = os.path.join(self.dir_val, self.name_task, "results.json")
        with open(path_out, "w") as f:
            json_metrics = ext_metrics(metrics)
            json.dump(json_metrics, f)
        # return mAP50
        return json_metrics["map50"]


def ext_metrics(metrics):
    """
    args
    ----
        metrics: dict
            metrics from model.val()

    return
    ------
        json: dict
    """
    # metrics
    map5095 = metrics.box.map.round(4)
    map50 = metrics.box.map50.round(4)
    precision = metrics.box.p[0].round(4)
    recall = metrics.box.r[0].round(4)
    f1 = metrics.box.f1[0].round(4)
    # confusion matrix
    conf_mat = metrics.confusion_matrix.matrix  # conf=0.25, iou_thres=0.45
    n_all = conf_mat[:, 0].sum()
    n_missed = conf_mat[1, 0].sum()
    n_false = conf_mat[0, 1].sum()
    # write json
    json_out = dict(
        map5095=map5095,
        map50=map50,
        precision=precision,
        recall=recall,
        f1=f1,
        n_all=int(n_all),
        n_missed=int(n_missed),
        n_false=int(n_false),
    )
    return json_out



from abc import ABC, abstractmethod


class Splitter(ABC):
    def __init__(
        self,
        root: str,
        ratio_train: float = 0.64,
        ratio_val: float = 0.16,
        ratio_test: float = 0.2,
    ):
        self.root = root
        self.config = None
        # ids
        self.ids = None
        self.id_train = None
        self.id_val = None
        self.id_test = None
        # split ratios
        self.ratio_train = ratio_train
        self.ratio_val = ratio_val
        self.ratio_test = ratio_test

    @abstractmethod
    def _set_dataset(self):
        """
        read or open the config file:
            COCO: JSON
            YOLO: YAML
        results:
            self.config
            self.ids
        """

    @abstractmethod
    def _get_ids(self) -> list:
        """
        get ids from the config file
        """

    def _set_ids(self) -> None:
        """
        set sample ids from the dataset to self.ids
        """
        self.ids = self._get_ids()

    @abstractmethod
    def shuffle_train_test(self) -> None:
        """
        shuffle ids into lists of train, val, test
        sets self.id_train, self.id_val, self.id_test
        """

    @abstractmethod
    def write_dataset(self) -> None:
        """
        write the images and configuration file based on the ids
        """



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="niche_cv",
    version="0.0.11",
    url="https://github.com/Niche-Squad/niche_cv",
    author="James Chen",
    author_email="niche@vt.edu",
    description="The Computer Vision Library for Niche Squad",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=[],
    entry_points={"console_scripts": ["niche=niche_cv.show:main"]},
)



"""
Take labels/_coco.json as input and split it into train, val, and test.
"""

from splitter.coco import COCO_Splitter
import os


def main():
    dir_root = os.path.join(os.path.dirname(__file__), "balloon", "labels")
    splitter = COCO_Splitter(dir_root=dir_root)
    splitter.shuffle_train_test()
    splitter.write_dataset()


if __name__ == "__main__":
    main()



import os
import json
import numpy as np
from torch.utils import data
from transformers import (
    DetrConfig,
    DetrModel,
    DetrForObjectDetection,
    DetrImageProcessor,
)

# local imports
from detection.data.loaders.coco import CocoDetectionDataModule
from detection.data.splitter.coco import CocoSplitter
from detection.models.detr import Detr

BACKBONE = "facebook/detr-resnet-50"
DIR_DATA = os.path.join(os.path.dirname(__file__), "detection", "data", "balloon")

# DETR fine-tuning note
# https://github.com/NielsRogge/Transformers-Tutorials/blob/master/DETR/Fine_tuning_DetrForObjectDetection_on_custom_dataset_(balloon).ipynb
# https://huggingface.co/docs/transformers/main/en/main_classes/configuration#transformers.PretrainedConfig

# shuffle the dataset
splitter = CocoSplitter(DIR_DATA)
splitter.shuffle_train_test()
splitter.write_dataset()

# load the dataset
processor = DetrImageProcessor.from_pretrained(BACKBONE)
data_module = CocoDetectionDataModule(DIR_DATA, processor, batch_size=4)
data_module.setup()
id2label = {k: v["name"] for k, v in data_module.train.coco.cats.items()}

# check the batch
batch = next(iter(data_module.train_dataloader()))
batch.keys()
# pixel_values, target = data_module.train[0]

# model
model = Detr(BACKBONE, len(id2label))
outputs = model(pixel_values=batch["pixel_values"], pixel_mask=batch["pixel_mask"])
outputs.logits.shape

from visualization.coco import show_dataset

show_dataset(train_dataset)

train_dataset.coco.getImgIds()


config = DetrConfig.from_pretrained(BACKBONE, num_labels=1)

model = DetrModel(config)
model.config.hidden_size

processor = DetrImageProcessor.from_pretrained("facebook/detr-resnet-50")

model.train()
